# coding=utf-8
"""
Base class for all arbitrage finder modes

(c) Copyright Bprotocol foundation 2023.
Licensed under MIT
"""
import abc
from typing import Any, Tuple, Dict, List, Union

import pandas as pd

from fastlane_bot.tools.cpc import T
from fastlane_bot.utils import num_format


class ArbitrageFinderBase:
    """
    Base class for all arbitrage finder modes
    """

    XS_ARBOPPS = "arbopps"
    XS_TI = "ti"
    XS_EXACT = "exact"
    XS_ORDSCAL = "ordscal"
    XS_AGGTI = "aggti"
    XS_ORDINFO = "ordinfo"
    XS_ENCTI = "encti"
    XS_ROUTE = "route"
    AM_REGULAR = "regular"
    AM_SINGLE = "single"
    AM_TRIANGLE = "triangle"
    AM_MULTI = "multi"
    AM_MULTI_TRIANGLE = "multi_triangle"
    AM_BANCOR_V3 = "bancor_v3"
    AO_TOKENS = "tokens"
    AO_CANDIDATES = "candidates"

    def __init__(
        self,
        flashloan_tokens,
        CCm,
        mode="bothin",
        result=AO_CANDIDATES,
        ConfigObj: Any = None,
        arb_mode: str = None,
    ):
        self.flashloan_tokens = flashloan_tokens
        self.CCm = CCm
        self.mode = mode
        self.result = result
        self.best_profit = 0
        self.best_src_token = None
        self.best_trade_instructions = None
        self.best_trade_instructions_df = None
        self.best_trade_instructions_dic = None
        self.ConfigObj = ConfigObj
        self.base_exchange = "bancor_v3" if arb_mode == "bancor_v3" else "carbon_v1"

    @abc.abstractmethod
    def find_arbitrage(self, candidates: List[Any] = None, ops: Tuple = None, best_profit: float = 0, profit_src: float = 0) -> Union[List, Tuple]:
        """
        See subclasses for details

        Parameters
        ----------
        candidates : List[Any], optional
            List of candidates, by default None
        ops : Tuple, optional
            Tuple of operations, by default None
        best_profit : float, optional
            Best profit so far, by default 0
        profit_src : float, optional
            Profit source, by default 0

        Returns
        -------
        Union[List, Tuple]
            If self.result == self.AO_CANDIDATES, it returns a list of candidates.
        """
        pass

    def _set_best_ops(
        self,
        best_profit: float,
        ops: Tuple,
        profit: float,
        src_token: str,
        trade_instructions: Any,
        trade_instructions_df: pd.DataFrame,
        trade_instructions_dic: Dict[str, Any],
    ) -> Tuple[float, Tuple]:
        """
        Set the best operations.

        Parameters:

        """
        self.ConfigObj.logger.debug("*************")
        self.ConfigObj.logger.debug(f"New best profit: {profit}")

        # Update the best profit and source token
        best_profit = profit
        best_src_token = src_token

        # Update the best trade instructions
        best_trade_instructions_df = trade_instructions_df
        best_trade_instructions_dic = trade_instructions_dic
        best_trade_instructions = trade_instructions

        self.ConfigObj.logger.debug(
            f"best_trade_instructions_df: {best_trade_instructions_df}"
        )

        # Update the optimal operations
        ops = (
            best_profit,
            best_trade_instructions_df,
            best_trade_instructions_dic,
            best_src_token,
            best_trade_instructions,
        )

        self.ConfigObj.logger.debug("*************")

        return best_profit, ops

    def calculate_profit(
        self, src_token: str, profit_src: float, CCm: Any, cids: List[str], profit: int = 0
    ) -> float:
        """
        Calculate profit based on the source token.
        """
        if src_token == T.BNT:
            profit = profit_src
        else:
            try:
                price_src_per_bnt = (
                    CCm.bypair(pair=f"{T.BNT}/{src_token}")
                    .byparams(exchange="bancor_v3")[0]
                    .p
                )
                profit = profit_src / price_src_per_bnt
            except Exception as e:
                self.ConfigObj.logger.error(f"[TODO CLEAN UP]{e}")
        self.ConfigObj.logger.debug(f"Profit in bnt: {num_format(profit)} {cids}")
        return profit

    @staticmethod
    def get_netchange(trade_instructions_df: pd.DataFrame) -> List[float]:
        """
        Get the net change from the trade instructions.
        """
        try:
            return trade_instructions_df.iloc[-1]
        except Exception:
            return [500]  # an arbitrary large number

    def handle_candidates(
        self,
        best_profit: float,
        profit: float,
        trade_instructions_df: pd.DataFrame,
        trade_instructions_dic: Dict[str, Any],
        src_token: str,
        trade_instructions: Any,
    ) -> List[Tuple[float, pd.DataFrame, Dict[str, Any], str, Any]]:
        """
        Handle candidate addition based on conditions.

        Parameters:
        ----------
        best_profit : float
            Best profit
        profit : float
            Profit
        trade_instructions_df : pd.DataFrame
            Trade instructions dataframe
        trade_instructions_dic : dict
            Trade instructions dictionary
        src_token : str
            Source token
        trade_instructions : any
            Trade instructions

        Returns:
        -------
        candidates : list
            Candidates
        """
        netchange = self.get_netchange(trade_instructions_df)
        condition_zeros_one_token = max(netchange) < 1e-4

        if (
            condition_zeros_one_token and profit > self.ConfigObj.DEFAULT_MIN_PROFIT
        ):  # candidate regardless if profitable
            return [
                (
                    profit,
                    trade_instructions_df,
                    trade_instructions_dic,
                    src_token,
                    trade_instructions,
                )
            ]
        return []

    def find_best_operations(
        self,
        best_profit: float,
        ops: Tuple[float, pd.DataFrame, Dict[str, Any], str, Any],
        profit: float,
        trade_instructions_df: pd.DataFrame,
        trade_instructions_dic: Dict[str, Any],
        src_token: str,
        trade_instructions: Any,
    ) -> Tuple[float, Tuple[float, pd.DataFrame, Dict[str, Any], str, Any]]:
        """
        Find the best operations based on conditions.

        Parameters:
        ----------
        best_profit : float
            Best profit
        ops : tuple
            Operations
        profit : float
            Profit
        trade_instructions_df : pd.DataFrame
            Trade instructions dataframe
        trade_instructions_dic : dict
            Trade instructions dictionary
        src_token : str
            Source token
        trade_instructions : any
            Trade instructions

        Returns:
        -------
        best_profit : float
            Best profit
        ops : tuple
            Operations
        """
        netchange = self.get_netchange(trade_instructions_df)
        condition_better_profit = profit > best_profit
        condition_zeros_one_token = max(netchange) < 1e-4
        if condition_better_profit and condition_zeros_one_token:
            return self._set_best_ops(
                best_profit,
                ops,
                profit,
                src_token,
                trade_instructions,
                trade_instructions_df,
                trade_instructions_dic,
            )
        return best_profit, ops



__VERSION__ = "1.2"
__DATE__="05/May/2023"

from typing import Dict, Any, List, Union

from _decimal import Decimal
from dataclasses import dataclass

#from fastlane_bot.config import SUPPORTED_EXCHANGES, CARBON_V1_NAME, UNISWAP_V3_NAME
from fastlane_bot.helpers.univ3calc import Univ3Calculator
from fastlane_bot.tools.cpc import ConstantProductCurve
from fastlane_bot.utils import UniV3Helper, EncodedOrder

from fastlane_bot.config import Config

@dataclass
class PoolAndTokens:
    """
    Represents a pool and its tokens in a single object. This is not a database model, but a helper class.

    Parameters
    ----------
    id : int
        The id of the pool
    cid : str
        The cid of the pool
    last_updated : str
        The last time the pool was updated
    last_updated_block : int
        The last block the pool was updated
    descr : str
        The description of the pool
    pair_name : str
        The name of the pool
    exchange_name : str
        The name of the exchange
    fee : Decimal
        The fee of the pool
    tkn0_balance : Decimal
        The balance of token 0
    tkn1_balance : Decimal
        The balance of token 1
    z_0 : Decimal
        The z_0 value
    y_0 : Decimal
        The y_0 value
    A_0 : Decimal
        The A_0 value
    B_0 : Decimal
        The B_0 value
    z_1 : Decimal
        The z_1 value
    y_1 : Decimal
        The y_1 value
    A_1 : Decimal
        The A_1 value
    B_1 : Decimal
        The B_1 value
    sqrt_price_q96 : Decimal
        The sqrt_price_q96 value
    tick : int
        The tick value
    tick_spacing : int
        The tick spacing value
    liquidity : Decimal
        The liquidity value
    address : str
        The address of the pool
    anchor : str
        The anchor of the pool
    tkn0 : str
        The token 0 key
    tkn1 : str
        The token 1 key
    tkn0_address : str
        The address of token 0
    tkn0_decimals : int
        The decimals of token 0
    tkn1_address : str
        The address of token 1
    tkn1_decimals : int
        The decimals of token 1

    """
    __VERSION__=__VERSION__
    __DATE__=__DATE__
    
    ConfigObj: Config
    id: int
    cid: str
    last_updated: str
    last_updated_block: int
    descr: str
    pair_name: str
    exchange_name: str
    fee: Decimal
    fee_float: float
    tkn0_balance: Decimal
    tkn1_balance: Decimal
    z_0: Decimal
    y_0: Decimal
    A_0: Decimal
    B_0: Decimal
    z_1: Decimal
    y_1: Decimal
    A_1: Decimal
    B_1: Decimal
    sqrt_price_q96: Decimal
    tick: int
    tick_spacing: int
    liquidity: Decimal
    address: str
    anchor: str
    tkn0: str
    tkn1: str
    tkn0_address: str
    tkn0_decimals: int
    tkn1_address: str
    tkn1_decimals: int
    tkn0_key: str = None
    tkn1_key: str = None
    ADDRDEC = None

    
    def __post_init__(self):
        self.tkn0_key = self.tkn0
        self.tkn1_key = self.tkn1
        self.A_1 = self.A_1 or 0
        self.B_1 = self.B_1 or 0
        self.A_0 = self.A_0 or 0
        self.B_0 = self.B_0 or 0
        self.z_0 = self.z_0 or 0
        self.y_0 = self.y_0 or 0
        self.z_1 = self.z_1 or 0
        self.y_1 = self.y_1 or 0
    
    def to_cpc(self) -> Union[ConstantProductCurve, List[Any]]:
        """
        converts self into an instance of the ConstantProductCurve class.
        """

        self.fee = float(Decimal(self.fee))
        if self.exchange_name == self.ConfigObj.UNISWAP_V3_NAME:
            out = self._univ3_to_cpc()
        elif self.exchange_name == self.ConfigObj.CARBON_V1_NAME:
            out = self._carbon_to_cpc()
        elif self.exchange_name in self.ConfigObj.SUPPORTED_EXCHANGES:
            out = self._other_to_cpc()
        else:
            raise NotImplementedError(f"Exchange {self.exchange_name} not implemented.")

        return out

    @property
    def _params(self):
        """
        creates the parameter dict for the ConstantProductCurve class
        """
        return {
                "exchange": str(self.exchange_name),
                "tknx_dec": int(self.tkn0_decimals),
                "tkny_dec": int(self.tkn1_decimals),
                "tknx_addr": str(self.tkn0_address),
                "tkny_addr": str(self.tkn1_address),
                "blocklud": int(self.last_updated_block),
        }
        
    def _other_to_cpc(self) -> List[Any]:
        """
        constructor: from Uniswap V2 pool (see class docstring for other parameters)

        :x_tknb:    current pool liquidity in token x (base token of the pair)*
        :y_tknq:    current pool liquidity in token y (quote token of the pair)*
        :k:         uniswap liquidity parameter (k = xy)*

        *exactly one of k,x,y must be None; all other parameters must not be None;
        a reminder that x is TKNB and y is TKNQ
        """
        
        # convert tkn0_balance and tkn1_balance to Decimal from wei
        tkn0_balance = self.convert_decimals(self.tkn0_balance, self.tkn0_decimals)
        tkn1_balance = self.convert_decimals(self.tkn1_balance, self.tkn1_decimals)

        # create a typed-dictionary of the arguments
        typed_args = {
            "x_tknb": tkn0_balance,
            "y_tknq": tkn1_balance,
            "pair": self.pair_name.replace("ETH-EEeE", "WETH-6Cc2"),
            "fee": self.fee,
            "cid": self.cid,
            "descr": self.descr,
            "params":  self._params,
        }
        return [ConstantProductCurve.from_univ2(**self._convert_to_float(typed_args))]

    class DoubleInvalidCurveError(ValueError): pass
    
    def _carbon_to_cpc(self) -> ConstantProductCurve:
        """
        constructor: from a single Carbon order (see class docstring for other parameters)*

        :yint:      current pool y-intercept**
        :y:         current pool liquidity in token y
        :pa:        fastlane_bot price range left bound (higher price in dy/dx)
        :pb:        fastlane_bot price range right bound (lower price in dy/dx)
        :A:         alternative to pa, pb: A = sqrt(pa) - sqrt(pb) in dy/dy
        :B:         alternative to pa, pb: B = sqrt(pb) in dy/dy
        :tkny:      token y
        :isdydx:    if True prices in dy/dx, if False in quote direction of the pair

        *Note that ALL parameters are mandatory, except that EITHER pa, bp OR A, B
        must be given but not both; we do not correct for incorrect assignment of
        pa and pb, so if pa <= pb IN THE DY/DX DIRECTION, MEANING THAT THE NUMBERS
        ENTERED MAY SHOW THE OPPOSITE RELATIONSHIP, then an exception will be raised

        **note that the result does not depend on yint, and for the time being we
        allow to omit yint (in which case it is set to y, but this does not make
        a difference for the result)
        """

        # if idx == 0, use the first curve, otherwise use the second curve. change the numerical values to Decimal
        lst = []
        errors = []
        for i in [0, 1]:
            #pair = self.pair_name.replace("ETH-EEeE", "WETH-6Cc2")
            S = Decimal(self.A_1) if i == 0 else Decimal(self.A_0)
            B = Decimal(self.B_1) if i == 0 else Decimal(self.B_0)
            y = Decimal(self.y_1) if i == 0 else Decimal(self.y_0)
            z = yint = Decimal(self.z_1) if i == 0 else Decimal(self.z_0)

            encoded_order = EncodedOrder(
                **{
                    "token": self.pair_name.split("/")[i].replace(
                        "ETH-EEeE", "WETH-6Cc2"
                    ),
                    "A": S,
                    "B": B,
                    "y": y,
                    "z": z,
                }
            )

            def decimal_converter(idx):
                if idx == 0:
                    dec0 = self.tkn0_decimals
                    dec1 = self.tkn1_decimals
                else:
                    dec0 = self.tkn1_decimals
                    dec1 = self.tkn0_decimals
                return Decimal(str(10 ** (dec0 - dec1)))

            decimal_converter = decimal_converter(i)

            p_start = Decimal(encoded_order.p_start) * decimal_converter
            p_end = Decimal(encoded_order.p_end) * decimal_converter
            yint = Decimal(yint) / (
                    Decimal("10") ** [self.tkn1_decimals, self.tkn0_decimals][i]
            )
            y = Decimal(y) / (
                    Decimal("10") ** [self.tkn1_decimals, self.tkn0_decimals][i]
            )

            tkny = 1 if i == 0 else 0
            typed_args = {
                "cid": f"{self.cid}-{i}",
                "yint": yint,
                "y": y,
                "pb": p_end,
                "pa": p_start,
                "tkny": self.pair_name.split("/")[tkny].replace(
                    "ETH-EEeE", "WETH-6Cc2"
                ),
                "pair": self.pair_name.replace("ETH-EEeE", "WETH-6Cc2"),
                "params": {"exchange": self.exchange_name},
                "fee": self.fee,
                "descr": self.descr,
                "params": self._params
            }
            try:
                if typed_args["y"] > 0:
                    lst.append(ConstantProductCurve.from_carbon(**self._convert_to_float(typed_args)))
                # else:
                #     self.ConfigObj.logger.debug(f"empty carbon pool [{typed_args['cid']}]")
            except Exception as e:
                errmsg = f"[_carbon_to_cpc] error in curve {i} [probably empty: {typed_args}] - [{e}]\n"
                self.ConfigObj.logger.debug(errmsg)
                errors += [errmsg]

        # if not len(lst) > 0:
        #     errmsg = f"[_carbon_to_cpc] error in BOTH curves {errors}\n\n"
        #     self.ConfigObj.logger.warning(errmsg)
        #     raise self.DoubleInvalidCurveError(errmsg)
            
        return lst
    
    FEE_LOOKUP = {
        0.0001: Univ3Calculator.FEE100,
        0.0005: Univ3Calculator.FEE500,
        0.0030: Univ3Calculator.FEE3000,
        0.01: Univ3Calculator.FEE10000,
    }
    def _univ3_to_cpc(self) -> List[Any]:
        """
        Preprocesses a Uniswap V3 pool params in order to create a ConstantProductCurve instance for optimization.

        :return: ConstantProductCurve
            :k:        pool constant k = xy [x=k/y, y=k/x]
            :x:        (virtual) pool state x (virtual number of base tokens for sale)
            :x_act:    actual pool state x (actual number of base tokens for sale)
            :y_act:    actual pool state y (actual number of quote tokens for sale)
            :pair:     tkn_address pair in slash notation ("TKNB/TKNQ"); TKNB is on the x-axis, TKNQ on the y-axis
            :cid:      unique id (optional)
            :fee:      fee (optional); eg 0.01 for 1%
            :descr:    description (optional; eg. "UniV3 0.1%")
            :params:   additional parameters (optional)
 
        """
        args = {"token0": self.tkn0_key, "token1": self.tkn1_key, "sqrt_price_q96": self.sqrt_price_q96, "tick": self.tick, "liquidity": self.liquidity}
        feeconst = self.FEE_LOOKUP.get(float(self.fee_float))
        if feeconst is None:
            raise ValueError(f"Illegal fee for Uniswap v3 pool: {self.fee_float} [{self.FEE_LOOKUP}]]")
        uni3 = Univ3Calculator.from_dict(args, feeconst, addrdec=self.ADDRDEC)
        params = uni3.cpc_params()
        #print("u3params", params)
        if params["uniL"] == 0:
            self.ConfigObj.logger.debug(f"empty univ3 pool [{self.cid}]")
            return []
        params["cid"] = self.cid
        params["descr"] = self.descr
        params["params"] = self._params
        return [ConstantProductCurve.from_univ3(**params)]

    @staticmethod
    def convert_decimals(tkn_balance_wei: Decimal, tkn_decimals: int) -> Decimal:
        """
        Converts wei balance to token balance (returns a Decimal)
        """
        # TODO: why Decimal?!?
        return Decimal(str(Decimal(str(tkn_balance_wei)) / (Decimal("10") ** Decimal(str(tkn_decimals)))))

    @staticmethod
    def _convert_to_float(typed_args: Dict[str, Any]) -> Dict[str, Any]:
        """
        converts Decimal to float
        """
        convert = lambda x: float(x) if isinstance(x, Decimal) else x
        return {k: convert(v) for k, v in typed_args.items()}
        # MIKE: this code converted the dict in place AND returned it; 
        # that may work but is ugly...
        # for key, value in typed_args.items():
        #     if isinstance(value, (Decimal, float)):
        #         typed_args[key] = float(value)
        # return typed_args



"""
Networks module for fastlane - used to interact with the blockchain.

(c) Copyright Bprotocol foundation 2023.
Licensed under MIT
"""


import subprocess
from abc import ABCMeta, ABC

from brownie import network, Contract, accounts  # type: ignore
from eth_typing import HexStr
from hexbytes import HexBytes
from web3 import Web3
from web3.types import TxReceipt

import os
from dotenv import load_dotenv
load_dotenv()

import logging
logger = logging.getLogger(__name__)


# *******************************************************************************************
# Singleton
# *******************************************************************************************


class Singleton(ABCMeta):
    """
    Singleton metaclass that enables the creation of a singleton object, as seen in this post:
    """

    _instances = {}

    def __call__(self, *args, **kwargs):
        if self not in self._instances:
            self._instances[self] = super(Singleton, self).__call__(*args, **kwargs)
        return self._instances[self]


# *******************************************************************************************
# Base Network
# *******************************************************************************************


class NetworkBase(ABC, metaclass=Singleton):
    """
    Base class for all networks - this is a singleton class that is used to interact with the blockchain

    :param network_id: the name of the network to connect to
    :param network_name: the name of the network to connect to
    :param provider_url: the url of the provider
    :param nonce: the nonce to use for transactions
    """

    web3: Web3

    def __init__(
        self,
        network_id: str,
        network_name: str,
        provider_url: str,
        provider_name: str,
        nonce: int,
    ):
        """
        :param provider_url: the RPC URL to connect to
        :param fastlane_contract_address: the address of the FastLane contract
        """
        self.network_id = network_id.lower()
        self.network_name = network_name
        self.provider_url = provider_url
        self.provider_name = provider_name
        self.web3 = Web3(Web3.HTTPProvider(provider_url))
        self.nonce = nonce


class EthereumNetwork(NetworkBase):
    """
    Ethereum network class

    :param network_id: the name of the network to connect to
    :param network_name: the name of the network to connect to
    :param provider_url: the url of the provider
    :param nonce: the nonce to use for transactions

    """

    chain_id: int = 1
    block_time: int = 0
    gas_price: int = 0
    gas_limit: int = 0
    _is_connected: bool = False

    def __init__(
        self,
        network_id: str = None,
        network_name: str = None,
        provider_url: str = None,
        provider_name: str = None,
        nonce: int = 0,
    ):
        """
        Note that Tenderly here must be configured in brownie - you can do this in the Terminal using the following command:
        brownie networks add [environment] [id] host=[host] [KEY=VALUE, ...]. For example:
        brownie networks add "Ethereum" "tenderly" host=https://rpc.tenderly.co/fork/7fd3f956-5409-4496-be95 chainid=1

        :param network_id: the name of the network to connect to
        :param network_name: the name of the network to connect to
        :param provider_url: the url of the provider
        :param provider_name: the name of the provider
        :param nonce: the nonce to use for transactions

        """

        super().__init__(
            network_id,
            network_name,
            provider_url,
            provider_name,
            nonce,
        )

    @property
    def is_connected(self) -> bool:
        """
        :return: True if the network is connected, False otherwise
        """
        return self._is_connected

    @is_connected.setter
    def is_connected(self, value: bool):
        self._is_connected = value

    def tx_to_hex(self, tx_receipt: TxReceipt) -> HexStr:
        """
        Convert a transaction receipt to a hex string
        :param tx_receipt:
        :return: the hex string
        """
        return self.web3.toHex(dict(tx_receipt)["transactionHash"])

    def sign_transaction(self, transaction: HexBytes) -> HexBytes:
        """
        Sign a transaction
        :param transaction:
        :return: the signed transaction
        """
        return self.web3.eth.account.sign_transaction(
            transaction, os.getenv("ETHERSCAN_TOKEN")
        )

    def increment_nonce(self) -> None:
        """
        increments nonce counter before submitting a transaction
        """
        self.nonce += 1

    def connect_network(self):
        """
        Connect to the network
        """
        logger.info(f"Connecting to {self.network_name} network")

        if self.is_connected:
            return

        # add_tenderly = f'brownie networks add "Ethereum" "{self.network_id}" host="{self.provider_url}"'
        # mod_tenderly = f'brownie networks modify "{self.network_id}" host="{self.provider_url}" name="{self.network_name}" chainid={self.chain_id}'
        # set_tenderly = f'brownie networks set_provider "{self.provider_name}"'

        # cmds = [add_tenderly, mod_tenderly, set_tenderly]
        # for cmd in cmds:
        #     p = subprocess.Popen(
        #         cmd,
        #         stdout=subprocess.PIPE,
        #         stderr=subprocess.PIPE,
        #         stdin=subprocess.PIPE,
        #         shell=True,
        #     )

        #     stdout, stderr = p.communicate()

        #     if "already exists" in stderr.decode("utf-8"):
        #         logger.debug(f"network {self.network_id} already exists")

        self.network = network
        self.network.connect(self.network_id)
        self._is_connected = True
        self.web3 = Web3(Web3.HTTPProvider(self.provider_url))
        logger.info(f"Connected to {self.network_id} network")
        logger.info(f"Connected to {self.web3.provider.endpoint_uri} network")



# coding=utf-8
"""
Base class for pairwise arbitrage finder modes

(c) Copyright Bprotocol foundation 2023.
Licensed under MIT
"""
import abc
import itertools
from typing import List, Tuple, Any, Union

from fastlane_bot.modes.base import ArbitrageFinderBase
from fastlane_bot.tools.cpc import CPCContainer


class ArbitrageFinderPairwiseBase(ArbitrageFinderBase):
    """
    Base class for pairwise arbitrage finder modes
    """

    @abc.abstractmethod
    def find_arbitrage(self, candidates: List[Any] = None, ops: Tuple = None, best_profit: float = 0, profit_src: float = 0) -> Union[List, Tuple]:
        """
        see base.py
        """
        pass

    @staticmethod
    def get_combos(
        CCm: CPCContainer, flashloan_tokens: List[str]
    ) -> Tuple[List[Any], List[Any]]:
        """
        Get combos for pairwise arbitrage

        Parameters
        ----------
        CCm : CPCContainer
            Container for all the curves
        flashloan_tokens : list
            List of flashloan tokens

        Returns
        -------
        all_tokens : list
            List of all tokens

        """
        all_tokens = CCm.tokens()
        flashloan_tokens_intersect = all_tokens.intersection(set(flashloan_tokens))
        combos = [
            (tkn0, tkn1)
            for tkn0, tkn1 in itertools.product(all_tokens, flashloan_tokens_intersect)
            # tkn1 is always the token being flash loaned
            # note that the pair is tkn0/tkn1, ie tkn1 is the quote token
            if tkn0 != tkn1
        ]
        return all_tokens, combos
